# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from asyncio import CancelledError
from collections import OrderedDict
import locale
import os
from pathlib import Path
import re
import sys
import traceback
import warnings

from colcon_core.environment_variable import EnvironmentVariable
from colcon_core.location import get_relative_package_index_path
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_grouped_by_priority
from colcon_core.plugin_system import SkipExtensionException
from colcon_core.prefix_path import get_chained_prefix_path
from colcon_core.subprocess import check_output

logger = colcon_logger.getChild(__name__)

"""Environment variable to enable all shell extensions."""
ALL_SHELLS_ENVIRONMENT_VARIABLE = EnvironmentVariable(
    'COLCON_ALL_SHELLS', 'Flag to enable all shell extensions')

use_all_shell_extensions = os.environ.get(
    ALL_SHELLS_ENVIRONMENT_VARIABLE.name, False)


class ShellExtensionPoint:
    """
    The interface for shell extensions.

    An shell extension generates the scripts for a specific shell to setup the
    environment.

    For each instance the attribute `SHELL_NAME` is being set to the basename
    of the entry point registering the extension.
    """

    """The version of the shell extension interface."""
    EXTENSION_POINT_VERSION = '2.1'

    """
    The default priority of shell extensions.

    A shell extension must use a higher priority than the default if and only
    if it is a "primary" shell.
    A "primary" shell does not depend on another shell to setup the
    environment, e.g. `sh`.
    An example for a "non-primary" shell would be `bash` which relies on the
    `sh` shell extension to setup environment variables and only contributes
    additional information like completion.

    All "non-primary" shell extensions must use a priority equal to or lower
    than the default.
    """
    PRIORITY = 100

    """
    The format string for a comment line.

    It must have the placeholder {comment}.
    This attribute must be defined in a subclass.
    """
    FORMAT_STR_COMMENT_LINE = None
    """
    The format string to set an environment variable.

    It must have the placeholder {name} and {value}.
    This attribute must be defined in a subclass.
    """
    FORMAT_STR_SET_ENV_VAR = None
    """
    The format string to use an environment variable.

    It must have the placeholder {name}.
    This attribute must be defined in a subclass.
    """
    FORMAT_STR_USE_ENV_VAR = None
    """
    The format string to invoke a script file.

    It must have the placeholder {prefix} and {script_path}.
    This attribute must be defined in a subclass.
    """
    FORMAT_STR_INVOKE_SCRIPT = None
    """
    The format string to remove a trailing separator.

    When prepending to an environment variable, a trailing separator will be
    left behind if the variable was not set previously.
    This command is used to cleanup the trailing separarator.

    It must have the placeholder {name} for the environment variable name.
    This attribute is optionally defined in subclasses.
    """
    FORMAT_STR_REMOVE_TRAILING_SEPARATOR = None

    def get_file_extensions(self):
        """
        Get the file extensions provided by this extension.

        By default the extension name will be returned.
        The method is intended to be overridden in a subclass.

        :returns: the file extensions
        :rtype: tuple
        """
        return (self.SHELL_NAME, )

    def create_prefix_script(self, prefix_path, merge_install):
        """
        Create a script in the install prefix path.

        The script should call each package specific script in order.

        This method must be overridden in a subclass.

        :param Path prefix_path: The path of the install prefix
        :param bool merge_install: The flag if all packages share the same
          install prefix
        """
        raise NotImplementedError()

    def _get_prefix_util_path(self):
        """
        Get the absolute path of the `prefix_util.py` module.

        :returns: The path of the module file
        :rtype: Path
        """
        warnings.warn(
            'colcon_core.shell._get_prefix_util_path() will be removed in the '
            'future', DeprecationWarning, stacklevel=2)
        return Path(__file__).parent / 'template' / 'prefix_util.py'

    def _get_prefix_util_template_path(self):
        """
        Get the absolute path of the `prefix_util.py.em` module template.

        :returns: The path of the module file
        :rtype: Path
        """
        return Path(__file__).parent / 'template' / 'prefix_util.py.em'

    def create_package_script(self, prefix_path, pkg_name, hooks):
        """
        Create a script for a specific package.

        The script should call each hook script in order.

        This method must be overridden in a subclass.

        :param Path prefix_path: The package specific install prefix
        :param str pkg_name: The package name
        :param list hooks: The relative paths to the hook scripts
        """
        raise NotImplementedError()

    def create_hook_set_value(
        self, env_hook_name, prefix_path, pkg_name, name, value,
    ):
        """
        Create a hook script to set an environment variable value.

        This method must be overridden in a subclass.

        :param str env_hook_name: The name of the hook script
        :param Path prefix_path: The path of the install prefix
        :param str pkg_name: The package name
        :param str name: The name of the environment variable
        :param str value: The value to be set
        :returns: The relative path to the created hook script
        :rtype: Path
        """
        raise NotImplementedError()

    def create_hook_append_value(
        self, env_hook_name, prefix_path, pkg_name, name, value,
    ):
        """
        Create a hook script to append a value to an environment variable.

        This method must be overridden in a subclass.

        :param str env_hook_name: The name of the hook script
        :param Path prefix_path: The path of the install prefix
        :param str pkg_name: The package name
        :param str name: The name of the environment variable
        :param str value: The value to be appended
        :returns: The relative path to the created hook script
        :rtype: Path
        """
        raise NotImplementedError()

    def create_hook_prepend_value(
        self, env_hook_name, prefix_path, pkg_name, name, subdirectory,
    ):
        """
        Create a hook script to prepend a value to an environment variable.

        This method must be overridden in a subclass.

        :param str env_hook_name: The name of the hook script
        :param Path prefix_path: The path of the install prefix
        :param str pkg_name: The package name
        :param str name: The name of the environment variable
        :param str subdirectory: The subdirectory of the prefix path
        :returns: The relative path to the created hook script
        :rtype: Path
        """
        raise NotImplementedError()

    def create_hook_include_file(
        self, env_hook_name, prefix_path, pkg_name, relative_path,
    ):
        """
        Create a hook script to include another script.

        This method must be overridden in a subclass.

        :param str env_hook_name: The name of the hook script
        :param Path prefix_path: The path of the install prefix
        :param str pkg_name: The package name
        :param str relative_path: The path of the included scripts
        :returns: The relative path to the created hook script
        :rtype: Path
        """
        raise NotImplementedError()

    async def generate_command_environment(
        self, task_name, build_base, dependencies,
    ):
        """
        Get the environment variables to invoke commands.

        The method must be overridden in a subclass if and only if the shell
        extension represents a "primary" shell (as defined in
        :attribute:`ShellExtensionPoint.PRIORITY`).

        :param str task_name: The name of the task
        :param Path build_base: The base path of the build directory
        :param set dependencies: The ordered dictionary mapping dependency
          names to their paths
        :returns: The environment
        :rtype: dict
        :raises SkipExtensionException: if the shell is not usable on the
          current platform
        """
        raise NotImplementedError()


def get_shell_extensions():
    """
    Get the available shell extensions.

    The extensions are grouped by their priority and each group is ordered by
    the entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.SHELL_NAME = name
    return order_extensions_grouped_by_priority(extensions)


async def get_command_environment(task_name, build_base, dependencies):
    """
    Get the environment variables to invoke commands.

    :param str task_name: The task name identifying a group of task extensions
    :param str build_base: The path of the build base
    :param dependencies: The ordered dictionary mapping dependency names to
      their paths
    """
    extensions = get_shell_extensions()
    for priority in extensions.keys():
        extensions_same_prio = extensions[priority]
        for extension in extensions_same_prio.values():
            try:
                # use the environment of the first successful shell extension
                return await extension.generate_command_environment(
                    task_name, Path(build_base), dependencies)
            except NotImplementedError:
                # skip extension, continue with next one
                logger.debug(
                    "Skip shell extension '{extension.SHELL_NAME}' for "
                    'command environment'.format_map(locals()))
            except SkipExtensionException as e:  # noqa: F841
                # skip extension, continue with next one
                logger.info(
                    "Skip shell extension '{extension.SHELL_NAME}' for "
                    'command environment: {e}'.format_map(locals()))
            except (CancelledError, RuntimeError):
                # re-raise same exception to handle it in the executor
                # without a traceback
                raise
            except Exception as e:  # noqa: F841
                # catch exceptions raised in shell extension
                exc = traceback.format_exc()
                logger.error(
                    'Exception in shell extension '
                    "'{extension.SHELL_NAME}': {e}\n{exc}"
                    .format_map(locals()))
                # skip failing extension, continue with next one
    raise RuntimeError(
        'Could not find a shell extension for the command environment')


async def get_environment_variables(cmd, *, cwd=None, shell=True):
    """
    Get the environment variables from the output of the command.

    :param args: the sequence of program arguments
    :param cwd: the working directory for the subprocess
    :param shell: whether to use the shell as the program to execute
    :rtype: dict
    """
    output = await check_output(cmd, cwd=cwd, shell=shell)
    env = OrderedDict()
    for line in output.splitlines():
        line = line.rstrip()
        if not line:
            continue
        encoding = locale.getpreferredencoding()
        try:
            line = line.decode(encoding)
        except UnicodeDecodeError:
            line_replaced = line.decode(encoding=encoding, errors='replace')
            logger.warning(
                'Failed to decode line from the environment using the '
                "encoding '{encoding}': {line_replaced}".format_map(locals()))
            continue
        parts = line.split('=', 1)
        if sys.platform != 'win32':
            regex = '^[a-zA-Z_][a-zA-Z0-9_]*$'
        else:
            regex = '^[a-zA-Z0-9%_' + ''.join(
                '\\' + c for c in r'(){}[]$*+-\/"#\',;.@!?'
            ) + ']+$'
        if len(parts) == 2 and re.match(regex, parts[0]):
            # add new environment variable
            env[parts[0]] = parts[1]
        else:
            # assume a line without an equal sign or with a "key" which is not
            # a valid name is a continuation of the previous line
            if env:
                env[list(env.keys())[-1]] += '\n' + line
    assert len(env) > 0, "The environment shouldn't be empty"
    return env


def create_environment_hook(
    env_hook_name, prefix_path, pkg_name, name, subdirectory, *, mode='prepend'
):
    """
    Create a hook script for each primary shell.

    :param str env_hook_name: The name of the hook script
    :param Path prefix_path: The path of the install prefix
    :param str pkg_name: The package name
    :param str name: The name of the environment variable
    :param str subdirectory: The value to be appended
    :param str mode: The mode how the new value should be combined with an
      existing value, currently only the value `prepend` is supported
    :returns: The relative paths to the created hook scripts
    :rtype: list
    """
    logger.log(
        1, "create_environment_hook('%s', '%s')" % (pkg_name, env_hook_name))

    hooks = []
    extensions = get_shell_extensions()
    for priority in extensions.keys():
        # only consider primary shell extensions
        if priority <= ShellExtensionPoint.PRIORITY:
            break

        extensions_same_prio = extensions[priority]
        for extension in extensions_same_prio.values():
            if mode == 'prepend':
                try:
                    hook = extension.create_hook_prepend_value(
                        env_hook_name, prefix_path, pkg_name, name,
                        subdirectory)
                    assert isinstance(hook, Path), \
                        'create_hook_prepend_value() should return a Path ' \
                        'object'

                except Exception as e:  # noqa: F841
                    # catch exceptions raised in shell extension
                    exc = traceback.format_exc()
                    logger.error(
                        'Exception in shell extension '
                        "'{extension.SHELL_NAME}': {e}\n{exc}"
                        .format_map(locals()))
                    # skip failing extension, continue with next one
                    continue
                hooks.append(hook)
            else:
                raise NotImplementedError()
    if not hooks:
        raise RuntimeError(
            'Could not find a primary shell extension for creating an '
            'environment hook')
    return hooks


_get_colcon_prefix_path_warnings = set()


def get_colcon_prefix_path(*, skip=None):
    """
    Get the paths from the COLCON_PREFIX_PATH environment variable.

    For not existing paths a warning is being printed and the path is being
    skipped.
    Even for repeated invocation a warning is only being printed once for each
    non existing path.

    :param skip: The current prefix path to be skipped and not be included in
      the return value
    :returns: The list of prefix paths
    :rtype: list
    """
    warnings.warn(
        "'colcon_core.shell.get_colcon_prefix_path()' has been deprecated, "
        "use 'colcon_core.prefix_path.get_chained_prefix_path() instead",
        stacklevel=2)
    global _get_colcon_prefix_path_warnings
    prefix_path = []
    colcon_prefix_path = os.environ.get('COLCON_PREFIX_PATH', '')
    for path in colcon_prefix_path.split(os.pathsep):
        if not path:
            continue
        if skip is not None and path == str(skip):
            continue
        if not os.path.exists(path):
            if path not in _get_colcon_prefix_path_warnings:
                logger.warning(
                    "The path '{path}' in the environment variable "
                    "COLCON_PREFIX_PATH doesn't exist".format_map(locals()))
            _get_colcon_prefix_path_warnings.add(path)
            continue
        prefix_path.append(path)

    return prefix_path


def check_dependency_availability(dependencies, *, script_filename):
    """
    Check if all dependencies are available.

    First the install base of the workspace is being checked.
    Second all prefix paths set in the environment are considered.
    In the second case a warning is logged to notify the user.

    :param dependencies: The ordered dictionary mapping dependency names to
      their paths. Packages which have been found in the environment are being
      removed from the dictionary.
    :param str script_filename: The filename of the package specific script to
      check for
    :raises RuntimeError: if any package isn't found in either of the locations
    """
    missing = OrderedDict()
    # check if the dependency exists in the install base of this workspace
    for pkg_name, pkg_install_base in dependencies.items():
        pkg_script = Path(
            pkg_install_base) / 'share' / pkg_name / script_filename
        if not pkg_script.exists():
            missing[pkg_name] = pkg_script

    # check if the dependency exists in any other prefix path
    packages_in_env = find_installed_packages_in_environment()
    env_packages = OrderedDict()
    for pkg_name, pkg_install_base in list(missing.items()):
        if pkg_name in packages_in_env:
            env_packages[pkg_name] = packages_in_env[pkg_name]
            # no need to source any script for this package
            del dependencies[pkg_name]
            del missing[pkg_name]

    # warn about using packages from the environment
    if env_packages:
        logger.warning(
            "The following packages are in the workspace but haven't been "
            'built:' +
            ''.join('\n- %s' % name for name in env_packages.keys()) +
            '\nThey are being used from the following locations instead:' +
            ''.join('\n- %s' % path for path in env_packages.values()) +
            '\nTo suppress this warning ignore these packages in the ' +
            'workspace:\n--packages-ignore ' + ' '.join(env_packages.keys()))

    # raise error in case any dependencies are not matched
    if missing:
        raise RuntimeError(
            'Failed to find the following files:' +
            ''.join('\n- %s' % path for path in missing.values()) +
            '\nCheck that the following packages have been built:' +
            ''.join('\n- %s' % name for name in missing.keys()))


def find_installed_packages_in_environment():
    """
    Find packages under the COLCON_PREFIX_PATH.

    For each prefix path the package index is being read and the first time a
    package is being found its install prefix is being added to the result.

    :returns: The mapping from a package name to the prefix path
    :rtype: OrderedDict
    """
    packages = OrderedDict()
    for prefix_path in get_chained_prefix_path():
        prefix_path = Path(prefix_path)
        pkgs = find_installed_packages(prefix_path)
        if pkgs is None:
            logger.debug(
                "Ignoring prefix path '{prefix_path}'".format_map(locals()))
            continue
        for pkg_name in sorted(pkgs.keys()):
            # ignore packages with the same name in "lower" prefix path
            if pkg_name in packages:
                continue
            packages[pkg_name] = pkgs[pkg_name]
    return packages


def find_installed_packages(install_base):
    """
    Find install packages under the install base path.

    The path must contain a marker file with the install layout.
    Based on the install layout the packages are discovered i different
    locations.

    :param Path install_base: The base path to find installed packages in
    :returns: The mapping from a package name to the prefix path, None if the
      path doesn't exist or doesn't contain a marker file with a valid install
      layout
    :rtype: OrderedDict or None
    """
    marker_file = install_base / '.colcon_install_layout'
    if not marker_file.is_file():
        return None
    install_layout = marker_file.read_text().rstrip()
    if install_layout not in ('isolated', 'merged'):
        return None

    packages = {}
    if install_layout == 'isolated':
        # for each subdirectory look for the package specific file
        for p in install_base.iterdir():
            if not p.is_dir():
                continue
            if p.name.startswith('.'):
                continue
            marker = p / get_relative_package_index_path() / p.name
            if marker.is_file():
                packages[p.name] = p
    else:
        # find all files in the subdirectory
        if (install_base / get_relative_package_index_path()).is_dir():
            package_index = install_base / get_relative_package_index_path()
            for p in package_index.iterdir():
                if not p.is_file():
                    continue
                if p.name.startswith('.'):
                    continue
                packages[p.name] = install_base
    return packages



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0


class JobQueued:
    """An event containing the identifier of the queued job."""

    __slots__ = ('identifier', 'dependencies')

    def __init__(self, identifier, dependencies=None):
        """
        Construct a JobQueued.

        :param str identifier: The job identifier
        :param set dependencies: The name of the recursive dependencies
        """
        self.identifier = identifier
        self.dependencies = dependencies


class JobStarted:
    """An event containing the identifier of the started job."""

    __slots__ = ('identifier', )

    def __init__(self, identifier):
        """
        Construct a JobStarted.

        :param str identifier: The job identifier
        """
        self.identifier = identifier


class JobProgress:
    """An event containing the identifier and progress of the job."""

    __slots__ = ('identifier', 'progress')

    def __init__(self, identifier, progress):
        """
        Construct a JobProgress.

        :param str identifier: The job identifier
        :param str progress: The message describing the progress of the job
        """
        self.identifier = identifier
        self.progress = progress


class JobEnded:
    """An event containing the identifier of the ended job and its rc."""

    __slots__ = ('identifier', 'rc')

    def __init__(self, identifier, rc):
        """
        Construct a JobEnded.

        :param str identifier: The job identifier
        :param rc: The return code of the job
        """
        self.identifier = identifier
        self.rc = rc


class JobSkipped:
    """An event containing the identifier of the skipped job."""

    __slots__ = ('identifier', )

    def __init__(self, identifier):
        """
        Construct a JobSkipped.

        :param str identifier: The job identifier
        """
        self.identifier = identifier



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from queue import Empty
from queue import Queue
from threading import Thread
import time
import traceback

from colcon_core.event.timer import TimerEvent
from colcon_core.event_handler import apply_event_handler_arguments
from colcon_core.event_handler import get_event_handler_extensions
from colcon_core.logging import colcon_logger

logger = colcon_logger.getChild(__name__)


class EventReactor:
    """Notify registered observers for events posted to the queue."""

    TIMER_INTERVAL = 0.1

    def __init__(self):  # noqa: D107
        self._thread = Thread(target=self._run)
        self._queue = Queue()
        self._observers = []
        self._last_timer_event = 0

    def get_queue(self):
        """Get the event queue."""
        return self._queue

    def register_observer(self, observer):
        """
        Register an observer which gets called for each event.

        :param callable observer: The callback
        """
        self._observers.append(observer)

    def _run(self):
        """
        Process events and notify all observers.

        If no events are being process for :py:attribute:`TIMER_INTERVAL`
        seconds a :class:`TimerEvent` is being generated and processed.

        An :class:`EventReactorShutdown` event will stop the loop.
        """
        while True:
            # send timer events in regular interval
            now = time.monotonic()
            time_since_last_timer_event = now - self._last_timer_event
            if time_since_last_timer_event >= self.TIMER_INTERVAL:
                self._notify_observers((TimerEvent(), None))
                self._last_timer_event = now
                timeout = self.TIMER_INTERVAL
            else:
                timeout = self.TIMER_INTERVAL - time_since_last_timer_event

            # wait for next event or timeout
            try:
                event = self._queue.get(timeout=timeout)
            except Empty:
                continue

            # publish event
            self._notify_observers(event)
            self._queue.task_done()

            # the signal to end the processing thread
            if len(event) > 1 and isinstance(event[0], EventReactorShutdown):
                break

    def _notify_observers(self, event):
        for observer in self._observers:
            try:
                retval = observer(event)
                assert retval is None, 'event handler should return None'
            except Exception as e:  # noqa: F841
                # catch exceptions raised in event handler extension
                exc = traceback.format_exc()
                logger.error(
                    'Exception in event handler extension '
                    "'{observer.EVENT_HANDLER_NAME}': {e}\n{exc}"
                    .format_map(locals()))
                # skip failing extension, continue with next one

    def flush(self):
        """Wait until the queue is empty."""
        while self._thread.is_alive():
            if self._queue.empty():
                return
            time.sleep(0.01)

    def start(self):
        """Start the event reactor."""
        self._thread.start()

    def stop(self):
        """
        Stop this event reactor and block until done.

        An :class:`EventReactorShutdown` event is added to the queue to notify
        all observers that the event reactor is shutting down.
        """
        self._queue.put((EventReactorShutdown(), None))
        logger.debug('joining thread')
        self._thread.join()
        logger.debug('joined thread')

    def __enter__(self):
        """
        Start the event reactor.

        :returns: self
        """
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Stop the event reactor."""
        self.stop()


class EventReactorShutdown:
    """An event generated before the event reactor is shut down."""

    __slots__ = ()


def create_event_reactor(context):
    """
    Create an event reactor and add all event handlers as observers.

    :param context: The context is passed to all event handlers
    :returns: The event reactor
    """
    event_reactor = EventReactor()
    event_handlers = get_event_handler_extensions(context=context)
    apply_event_handler_arguments(event_handlers, context.args)

    # register enabled event handlers
    for event_handler in event_handlers.values():
        if not event_handler.enabled:
            continue
        event_reactor.register_observer(event_handler)

    return event_reactor



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.package_identification.python \
    import create_dependency_descriptor
from colcon_core.package_identification.python \
    import PythonPackageIdentification
import pytest


def test_identify():
    extension = PythonPackageIdentification()

    with TemporaryDirectory(prefix='test_colcon_') as basepath:
        desc = PackageDescriptor(basepath)
        desc.type = 'other'
        assert extension.identify(desc) is None
        assert desc.name is None

        desc.type = None
        assert extension.identify(desc) is None
        assert desc.name is None
        assert desc.type is None

        basepath = Path(basepath)
        (basepath / 'setup.py').write_text('setup()')
        assert extension.identify(desc) is None
        assert desc.name is None
        assert desc.type is None

        (basepath / 'setup.cfg').write_text('')
        assert extension.identify(desc) is None
        assert desc.name is None
        assert desc.type is None

        (basepath / 'setup.cfg').write_text(
            '[metadata]\n'
            'name = pkg-name\n')
        assert extension.identify(desc) is None
        assert desc.name == 'pkg-name'
        assert desc.type == 'python'

        desc.name = 'other-name'
        with pytest.raises(RuntimeError) as e:
            extension.identify(desc)
        assert str(e.value).endswith(
            'Package name already set to different value')

        (basepath / 'setup.cfg').write_text(
            '[metadata]\n'
            'name = other-name\n'
            '[options]\n'
            'setup_requires =\n'
            "  build; sys_platform != 'win32'\n"
            "  build-windows; sys_platform == 'win32'\n"
            'install_requires =\n'
            '  runA > 1.2.3\n'
            '  runB\n'
            'tests_require = test == 2.0.0\n'
            'zip_safe = false\n')
        assert extension.identify(desc) is None
        assert desc.name == 'other-name'
        assert desc.type == 'python'
        assert set(desc.dependencies.keys()) == {'build', 'run', 'test'}
        assert desc.dependencies['build'] == {'build', 'build-windows'}
        assert desc.dependencies['run'] == {'runA', 'runB'}
        dep = next(x for x in desc.dependencies['run'] if x == 'runA')
        assert dep.metadata['version_gt'] == '1.2.3'
        assert desc.dependencies['test'] == {'test'}

        assert callable(desc.metadata['get_python_setup_options'])
        options = desc.metadata['get_python_setup_options'](None)
        assert 'zip_safe' in options


def test_create_dependency_descriptor():
    eq_str = 'pkgname==2.2.0'
    dep = create_dependency_descriptor(eq_str)
    assert dep.metadata['version_eq'] == '2.2.0'

    lt_str = 'pkgname<2.3.0'
    dep = create_dependency_descriptor(lt_str)
    assert dep.metadata['version_lt'] == '2.3.0'

    lte_str = 'pkgname<=2.2.0'
    dep = create_dependency_descriptor(lte_str)
    assert dep.metadata['version_lte'] == '2.2.0'

    gt_str = 'pkgname>2.3.0'
    dep = create_dependency_descriptor(gt_str)
    assert dep.metadata['version_gt'] == '2.3.0'

    gte_str = 'pkgname>=2.2.0'
    dep = create_dependency_descriptor(gte_str)
    assert dep.metadata['version_gte'] == '2.2.0'

    neq_str = 'pkgname!=1.2.1'
    dep = create_dependency_descriptor(neq_str)
    assert dep.metadata['version_neq'] == '1.2.1'

    compat_str = 'pkgname~=1.4.1a4'
    dep = create_dependency_descriptor(compat_str)
    assert dep.metadata['version_gte'] == '1.4.1a4'
    assert dep.metadata['version_lt'] == '1.5'

    compat_str = 'pkgname~=1.4.1'
    dep = create_dependency_descriptor(compat_str)
    assert dep.metadata['version_gte'] == '1.4.1'
    assert dep.metadata['version_lt'] == '1.5'

    compat_str = 'pkgname~=1.4.1.4'
    dep = create_dependency_descriptor(compat_str)
    assert dep.metadata['version_gte'] == '1.4.1.4'
    assert dep.metadata['version_lt'] == '1.4.2'

    compat_str = 'pkgname~=11.12'
    dep = create_dependency_descriptor(compat_str)
    assert dep.metadata['version_gte'] == '11.12'
    assert dep.metadata['version_lt'] == '12.0'

    multi_str = 'pkgname<=3.2.0, >=2.2.0'
    dep = create_dependency_descriptor(multi_str)
    assert dep.metadata['version_gte'] == '2.2.0'
    assert dep.metadata['version_lte'] == '3.2.0'



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.package_descriptor import PackageDescriptor


class PackageDecorator:
    """Decorator of a package descriptor to collect recursive dependencies."""

    __slots__ = (
        'descriptor',
        'recursive_dependencies',
        'selected',
    )

    def __init__(self, descriptor: PackageDescriptor):
        """
        Decorate a package descriptor.

        The function :function:`add_recursive_dependencies` should be invoked
        to populate the recursive dependencies.

        :param descriptor: The package descriptor
        """
        self.descriptor = descriptor
        self.recursive_dependencies = None
        self.selected = True


def get_decorators(descriptors):
    """
    Get decorators for package descriptors.

    :param Iterable descriptors: The package descriptors
    :returns: The package decorators
    :rtype: list
    """
    return [PackageDecorator(d) for d in descriptors]


def add_recursive_dependencies(
    decorators, direct_categories=None, recursive_categories=None,
):
    """
    Update the recursive dependencies of the decorators.

    :param set decorators: The known packages to consider
    :param Iterable[str] direct_categories: The names of the direct categories
    :param Iterable[str] recursive_categories: The names of the recursive
    categories
    """
    descriptors = [decorator.descriptor for decorator in decorators]
    for decorator in decorators:
        decorator.recursive_dependencies = \
            decorator.descriptor.get_recursive_dependencies(
                descriptors,
                direct_categories=direct_categories,
                recursive_categories=recursive_categories)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.environment import create_environment_hooks
from colcon_core.environment import create_environment_scripts
from colcon_core.environment import EnvironmentExtensionPoint
from colcon_core.environment import get_environment_extensions
from colcon_core.shell import get_shell_extensions
from colcon_core.shell import ShellExtensionPoint
from mock import Mock
from mock import patch
import pytest

from .entry_point_context import EntryPointContext


def test_extension_interface():
    extension = EnvironmentExtensionPoint()
    with pytest.raises(NotImplementedError):
        extension.create_environment_hooks(None, None)


class Extension1(EnvironmentExtensionPoint):

    def create_environment_hooks(self, prefix_path, pkg_name):
        return [
            '{prefix_path}/share/{pkg_name}/hook/one.ext'.format_map(locals()),
            '{prefix_path}/share/{pkg_name}/hook/two.ext'.format_map(locals()),
        ]


class Extension2(EnvironmentExtensionPoint):
    PRIORITY = 110


def test_get_environment_extensions():
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_environment_extensions()
    assert list(extensions.keys()) == ['extension2', 'extension1']


class Extension3(ShellExtensionPoint):
    pass


class Extension4(ShellExtensionPoint):
    pass


def test_create_environment_scripts():
    with TemporaryDirectory(prefix='test_colcon_') as basepath:
        pkg = Mock()
        pkg.name = 'name'
        pkg.dependencies = {}
        pkg.hooks = []
        args = Mock()
        args.install_base = basepath

        # no hooks at all
        with patch(
            'colcon_core.environment.create_environment_hooks', return_value=[]
        ):
            with patch(
                'colcon_core.environment.get_shell_extensions', return_value={}
            ):
                create_environment_scripts(pkg, args)

        pkg.hooks = [os.path.join(basepath, 'subA')]
        with EntryPointContext(extension3=Extension3, extension4=Extension4):
            extensions = get_shell_extensions()
            # one invalid return value, one check correct hooks argument
            extensions[100]['extension3'].create_package_script = Mock()
            extensions[100]['extension4'].create_package_script = Mock(
                return_value=None)
            with patch('colcon_core.environment.logger.error') as error:
                create_environment_scripts(
                    pkg, args, default_hooks=[('subB', )],
                    additional_hooks=[['subC', 'arg1', 'arg2']])
            # the raised exception is catched and results in an error message
            assert error.call_count == 1
            assert len(error.call_args[0]) == 1
            assert error.call_args[0][0].startswith(
                "Exception in shell extension 'extension3': "
                'create_package_script() should return None\n')
            # check for correct hooks argument
            mock = extensions[100]['extension4'].create_package_script
            assert mock.call_count == 1
            assert len(mock.call_args[0]) == 3
            assert mock.call_args[0][0] == Path(args.install_base)
            assert mock.call_args[0][1] == pkg.name
            hook_tuples = mock.call_args[0][2]
            assert len(hook_tuples) == 3
            assert hook_tuples[0] == ('subB', ())
            assert hook_tuples[1] == ('subC', ['arg1', 'arg2'])
            assert hook_tuples[2] == ('subA', [])


def test_create_environment_hooks():
    with TemporaryDirectory(prefix='test_colcon_') as basepath:
        with EntryPointContext(extension1=Extension1, extension2=Extension2):
            with patch('colcon_core.environment.logger.error') as error:
                hooks = create_environment_hooks(basepath, 'pkg_name')
    assert len(hooks) == 2
    assert hooks[0] == '{basepath}/share/pkg_name/hook/one.ext' \
        .format_map(locals())
    assert hooks[1] == '{basepath}/share/pkg_name/hook/two.ext' \
        .format_map(locals())
    # the raised exception is catched and results in an error message
    assert error.call_count == 1
    assert len(error.call_args[0]) == 1
    assert error.call_args[0][0].startswith(
        "Exception in environment extension 'extension2': \n")



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import OrderedDict

from colcon_core.argument_parser import ArgumentParserDecorator


class DestinationCollectorDecorator(ArgumentParserDecorator):
    """Collect the option names and destination of arguments."""

    def __init__(self, parser, **kwargs):
        """
        Construct a DestinationCollectorDecorator.

        :param parser: The argument parser to decorate
        """
        # avoid setting members directly, the base class overrides __setattr__
        # pass them as keyword arguments instead
        super().__init__(
            parser,
            _destinations=OrderedDict(),
            **kwargs)

    def get_destinations(self):
        """
        Get destinations for all added arguments.

        :returns: The destination names
        :rtype: OrderedDict
        """
        destinations = OrderedDict()
        destinations.update(self._destinations)
        for d in self._nested_decorators:
            destinations.update(d.get_destinations())
        return destinations

    def add_argument(self, *args, **kwargs):
        """Collect option names and destination for all added arguments."""
        argument = self._parser.add_argument(*args, **kwargs)

        for arg in args:
            if not arg.startswith('-'):
                continue
            key = arg.lstrip('-')
            assert key not in self._destinations
            self._destinations[key] = argument.dest

        return argument



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import copy
import logging
import os
from pathlib import Path

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_name

logger = colcon_logger.getChild(__name__)


class VerbExtensionPoint:
    """
    The interface for verb extensions.

    A verb extension provides a verb to the command line tool.

    For each instance the attribute `VERB_NAME` is being set to the basename of
    the entry point registering the extension.
    """

    """The version of the verb extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    def add_arguments(self, *, parser):
        """
        Add command line arguments specific to the verb.

        The method is intended to be overridden in a subclass.

        :param parser: The argument parser
        """
        pass

    def main(self, *, context):
        """
        Execute the verb extension logic.

        This method must be overridden in a subclass.

        :param context: The context providing the parsed command line arguments
        :returns: The return code
        """
        raise NotImplementedError()


def get_verb_extensions():
    """
    Get the available verb extensions.

    The extensions are ordered by their entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.VERB_NAME = name
    return order_extensions_by_name(extensions)


def check_and_mark_build_tool(build_base, *, this_build_tool='colcon'):
    """
    Check the marker file for the previous build tool, otherwise create it.

    The marker filename is `.built_by`.

    :param str build_base: The build directory
    :param str this_build_tool: The name of this build tool
    :raises RuntimeError: if the marker file contains the name of a different
      build tool
    """
    marker_path = Path(build_base) / '.built_by'
    if marker_path.parent.is_dir():
        if marker_path.is_file():
            previous_build_tool = marker_path.read_text().rstrip()
            if previous_build_tool == this_build_tool:
                return
            raise RuntimeError(
                "The build directory '{build_base}' was created by "
                "'{previous_build_tool}'. Please remove the build directory "
                'or pick a different one.'.format_map(locals()))
    else:
        os.makedirs(build_base, exist_ok=True)

    marker_path.write_text(this_build_tool + '\n')


def check_and_mark_install_layout(install_base, *, merge_install):
    """
    Check the marker file for the previous install layout, otherwise create it.

    The marker filename is `.colcon_install_layout`.

    :param str install_base: The install directory
    :param bool merge_install: The flag if all packages share the same prefix
    :raises RuntimeError: if the marker file contains a different install
      layout
    """
    this_install_layout = 'merged' if merge_install else 'isolated'
    marker_path = Path(install_base) / '.colcon_install_layout'
    if marker_path.parent.is_dir():
        if marker_path.is_file():
            previous_install_layout = marker_path.read_text().rstrip()
            if previous_install_layout == this_install_layout:
                return
            change_option = 'remove' if merge_install else 'add'
            raise RuntimeError(
                "The install directory '{install_base}' was created with the "
                "layout '{previous_install_layout}'. Please remove the "
                'install directory, pick a different one or {change_option} '
                "the '--merge-install' option.".format_map(locals()))
    else:
        try:
            os.makedirs(install_base, exist_ok=True)
        except FileExistsError:
            raise RuntimeError(
                "The install base '{install_base}' is not a directory"
                .format_map(locals()))

    marker_path.write_text(this_install_layout + '\n')


def update_object(
    object_, key, value, package_name, argument_type, value_source
):
    """
    Set or update an attribute of an object.

    If the attribute exists and the passed value as well as the current value
    of the attribute are dictionaries then the current values are being updated
    with the passed values.

    If the attribute exists and the passed value as well as the current value
    of the attribute are lists then the passed values are being appended to the
    current values.

    Otherwise the attribute is being set to the passed value potentially
    overwriting an existing value.

    :param key: The name of the attributes
    :param value: The value used to set or update the attribute
    :param str package_name: The package name, only used for log messages
    :param str argument_type: The argument type, only used in log messages
    :param str value_source: The source of the value, only used for log
      messages
    """
    if not hasattr(object_, key):
        logger.log(
            5, "set package '{package_name}' {argument_type} argument '{key}' "
            "from {value_source} to '{value}'".format_map(locals()))
        # add value to the object
        # copy value to avoid changes to either of them to affect each other
        setattr(object_, key, copy.deepcopy(value))
        return

    old_value = getattr(object_, key)
    if isinstance(old_value, dict) and isinstance(value, dict):
        logger.log(
            5, "update package '{package_name}' {argument_type} argument "
            "'{key}' from {value_source} with '{value}'".format_map(locals()))
        # update dictionary
        old_value.update(value)
        return

    if isinstance(old_value, list) and isinstance(value, list):
        logger.log(
            5, "extend package '{package_name}' {argument_type} argument "
            "'{key}' from {value_source} with '{value}'".format_map(locals()))
        # extend list
        old_value += value
        return

    severity = 5 \
        if old_value is None or type(old_value) == type(value) \
        else logging.WARNING
    logger.log(
        severity, "overwrite package '{package_name}' {argument_type} "
        "argument '{key}' from {value_source} with '{value}' (before: "
        "'{old_value}')".format_map(locals()))
    # overwrite existing value
    # copy value to avoid changes to either of them to affect each other
    setattr(object_, key, copy.deepcopy(value))



# Copyright 2010 Dirk Thomas
# Licensed under the Apache License, Version 2.0

# This module exists for multi-Python environments
# so the user can invoke colcon as `python -m colcon`

import sys

import colcon_core.command

sys.exit(colcon_core.command.main())



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.package_identification import _identify
from colcon_core.package_identification \
    import get_package_identification_extensions
from colcon_core.package_identification import identify
from colcon_core.package_identification import IgnoreLocationException
from colcon_core.package_identification \
    import PackageIdentificationExtensionPoint
from mock import Mock
from mock import patch
import pytest

from .entry_point_context import EntryPointContext


class Extension1(PackageIdentificationExtensionPoint):
    PRIORITY = 80


class Extension2(PackageIdentificationExtensionPoint):
    pass


class Extension3(PackageIdentificationExtensionPoint):
    PRIORITY = 90


class Extension4(PackageIdentificationExtensionPoint):
    pass


def test_get_package_identification_extensions():
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2,
        extension3=Extension3, extension4=Extension4,
    ):
        extensions = get_package_identification_extensions()
    assert list(extensions.keys()) == [100, 90, 80]
    assert list(extensions[100].keys()) == ['extension2', 'extension4']
    assert list(extensions[90].keys()) == ['extension3']
    assert list(extensions[80].keys()) == ['extension1']


def identify_name(desc):
    desc.name = 'name'


def identify_type(desc):
    desc.type = 'type'


def identify_name_and_type(desc):
    identify_name(desc)
    identify_type(desc)


def test_identify():
    path = '/some/path'
    context = EntryPointContext(
        extension1=Extension1, extension2=Extension2,
        extension3=Extension3, extension4=Extension4)

    with context:
        # no identification
        desc = identify({}, path)
        assert desc is None

        # no complete identification
        extensions = get_package_identification_extensions()
        extensions[80]['extension1'].identify = Mock(
            side_effect=identify_name)
        desc = identify(extensions, path)
        assert desc is None

        # valid result combined across priority groups
        extensions = get_package_identification_extensions()
        extensions[100]['extension4'].identify = Mock(
            side_effect=identify_type)
        desc = identify(extensions, path)
        assert isinstance(desc, PackageDescriptor)
        assert str(desc.path) == '/some/path'.replace('/', os.sep)
        assert desc.name == 'name'
        assert desc.type == 'type'

        # skip location
        extensions = get_package_identification_extensions()
        extensions[90]['extension3'].identify = Mock(
            side_effect=IgnoreLocationException())
        with pytest.raises(IgnoreLocationException):
            identify(extensions, path)

        # valid result from first priority group
        # lower priority groups are not even invoked
        extensions = get_package_identification_extensions()
        extensions[100]['extension4'].identify.side_effect = \
            identify_name_and_type
        desc = identify(extensions, path)
        assert isinstance(desc, PackageDescriptor)
        assert str(desc.path) == '/some/path'.replace('/', os.sep)
        assert desc.name == 'name'
        assert desc.type == 'type'

    with context:
        # multiple different results result in skipping the location
        extensions = get_package_identification_extensions()
        extensions[100]['extension2'].identify = Mock(
            side_effect=identify_name)
        extensions[100]['extension4'].identify = Mock(
            side_effect=identify_type)
        with pytest.raises(IgnoreLocationException):
            identify(extensions, path)


def test__identify():
    desc_path_only = PackageDescriptor('/some/path')
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2,
        extension3=Extension3, extension4=Extension4,
    ):
        # valid result
        extensions = get_package_identification_extensions()[100]
        extensions['extension2'].identify = Mock()
        extensions['extension4'].identify = identify_name_and_type
        desc = _identify(extensions, desc_path_only)
        assert isinstance(desc, PackageDescriptor)
        assert str(desc.path) == '/some/path'.replace('/', os.sep)
        assert desc.name == 'name'
        assert desc.type == 'type'

        # no results
        extensions = get_package_identification_extensions()[100]
        extensions['extension2'].identify = Mock()
        extensions['extension4'].identify = Mock()
        desc = _identify(extensions, desc_path_only)
        assert desc is None

        # multiple different results
        extensions = get_package_identification_extensions()[100]
        extensions['extension2'].identify = identify_name
        extensions['extension4'].identify = identify_type
        with patch(
            'colcon_core.package_identification.logger.warning'
        ) as warn:
            desc = _identify(extensions, desc_path_only)
            assert desc is False
            # the raised exception is catched and results in a warn message
            assert warn.call_count == 1
            assert len(warn.call_args[0]) == 1
            assert 'multiple matches' in warn.call_args[0][0]

        # invalid return value
        extensions = get_package_identification_extensions()[90]
        extensions['extension3'].identify = Mock(return_value=True)
        with patch('colcon_core.package_identification.logger.error') as error:
            desc = _identify(extensions, desc_path_only)
            assert desc is None
            # the raised exception is catched and results in an error message
            assert error.call_count == 1
            assert len(error.call_args[0]) == 1
            assert error.call_args[0][0].startswith(
                "Exception in package identification extension 'extension3' "
                "in '/some/path': identify() should return None\n"
                .replace('/', os.sep))

        # skip location
        extensions = get_package_identification_extensions()[90]
        extensions['extension3'].identify = Mock(
            side_effect=IgnoreLocationException())
        with pytest.raises(IgnoreLocationException):
            _identify(extensions, desc_path_only)

        # raise exception
        extensions = get_package_identification_extensions()[90]
        extensions['extension3'].identify = Mock(
            side_effect=RuntimeError('custom exception'))
        with patch('colcon_core.package_identification.logger.error') as error:
            desc = _identify(extensions, desc_path_only)
            assert desc is None
            # the raised exception is catched and results in an error message
            assert error.call_count == 1
            assert len(error.call_args[0]) == 1
            assert error.call_args[0][0].startswith(
                "Exception in package identification extension 'extension3' "
                "in '/some/path': custom exception\n".replace('/', os.sep))
